#include "framework.h"

#pragma comment(lib, "d3d9.lib")
#pragma comment(lib, "d3dx9.lib")
#pragma comment(lib, "winmm.lib")

#include <windows.h>
#include <d3d9.h>
#include <d3dx9.h>

#define SAFE_RELEASE(x) if(x != NULL){ x->Release(); x = NULL; }

#define APPNAME L"d3dproject"
#define CLASSNAME L"DIRECT3D"
bool isProcess = true;
RECT WindowRect;
int Width = 320, Height = 240;
bool fullscreen = false;
bool keys[256];
HWND hWnd;

LPDIRECT3D9 g_pD3D = NULL;
LPDIRECT3DDEVICE9 g_pd3dDevice = NULL;
D3DPRESENT_PARAMETERS d3dpp;

LPDIRECT3DVERTEXBUFFER9 MyVertexBuffer;
LPDIRECT3DTEXTURE9 MyTexture;

struct MyVertex {
    float x, y, z;
    DWORD color;
    float tx, ty;
};

DWORD MyVertex_FVF = D3DFVF_XYZ | D3DFVF_DIFFUSE | D3DFVF_TEX1;



void CenterWindow(HWND hWnd) {
    //    
    int screenWidth = GetSystemMetrics(SM_CXSCREEN);
    int screenHeight = GetSystemMetrics(SM_CYSCREEN);

    //   
    RECT windowRect;
    GetWindowRect(hWnd, &windowRect);
    int windowWidth = windowRect.right - windowRect.left;
    int windowHeight = windowRect.bottom - windowRect.top;

    //       
    int posX = (screenWidth - windowWidth) / 2;
    int posY = (screenHeight - windowHeight) / 2;

    //     
    SetWindowPos(hWnd, NULL, posX, posY, 0, 0, SWP_NOZORDER | SWP_NOSIZE);
}

void ReSizeD3DScene(int width, int height) {
    if (height == 0)
        height = 1;

    D3DXMATRIXA16 matProjection;
    D3DXMatrixPerspectiveFovLH(&matProjection, 45.0f, (float)width / (float)height, 0.1f, 100.0f);

    if (g_pd3dDevice)
        g_pd3dDevice->SetTransform(D3DTS_PROJECTION, &matProjection);
}

bool Cleanup() {
    return true;
}

bool ReInit() {
    ReSizeD3DScene(Width, Height);

    g_pd3dDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
    g_pd3dDevice->SetRenderState(D3DRS_LIGHTING, FALSE);
    g_pd3dDevice->SetRenderState(D3DRS_ZENABLE, TRUE);

    return true;
}

bool InitD3D(int width, int height, bool fscreen) {
    if (NULL == (g_pD3D = Direct3DCreate9(D3D_SDK_VERSION)))
        MessageBox(NULL, L"   Direct3D9!", L"", MB_OK | MB_ICONEXCLAMATION);

    ZeroMemory(&d3dpp, sizeof(d3dpp));

    if (fscreen) {
        d3dpp.BackBufferWidth = width;
        d3dpp.BackBufferHeight = height;
        d3dpp.BackBufferFormat = D3DFMT_X8R8G8B8;
        d3dpp.BackBufferCount = 1;
        d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
        d3dpp.hDeviceWindow = hWnd;
        d3dpp.Windowed = FALSE;
        d3dpp.EnableAutoDepthStencil = TRUE;
        d3dpp.AutoDepthStencilFormat = D3DFMT_D16;
        d3dpp.FullScreen_RefreshRateInHz = D3DPRESENT_RATE_DEFAULT;
        d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_DEFAULT;
    }
    else {
        d3dpp.Windowed = TRUE;
        d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
        d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;
        d3dpp.EnableAutoDepthStencil = TRUE;
        d3dpp.AutoDepthStencilFormat = D3DFMT_D16;
        d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_DEFAULT;
    }

    if (FAILED(g_pD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd,
        D3DCREATE_SOFTWARE_VERTEXPROCESSING, &d3dpp, &g_pd3dDevice))) {

        if (FAILED(g_pD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_REF, hWnd,
            D3DCREATE_SOFTWARE_VERTEXPROCESSING, &d3dpp, &g_pd3dDevice))) {

            MessageBox(NULL, L"    !", L"", MB_OK | MB_ICONEXCLAMATION);
            return false;
        }
    }

    return true;
}

bool ResetWindow(int width, int height, bool fscreen) {
    DWORD WindowStyle, WindowExStyle;

    if (fscreen) {
        WindowExStyle = WS_EX_APPWINDOW;
        WindowStyle = WS_POPUP;
    }
    else {
        WindowExStyle = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
        WindowStyle = WS_OVERLAPPEDWINDOW;
    }

    WindowRect.left = (long)0;
    WindowRect.right = (long)Width;
    WindowRect.top = (long)0;
    WindowRect.bottom = (long)Height;

    AdjustWindowRectEx(&WindowRect, WindowStyle, FALSE, WindowExStyle);

    ZeroMemory(&d3dpp, sizeof(d3dpp));

    if (fscreen) {
        d3dpp.BackBufferWidth = Width;
        d3dpp.BackBufferHeight = Height;
        d3dpp.BackBufferFormat = D3DFMT_X8R8G8B8;
        d3dpp.BackBufferCount = 1;
        d3dpp.FullScreen_RefreshRateInHz = D3DPRESENT_RATE_DEFAULT;
    }
    else {
        d3dpp.FullScreen_RefreshRateInHz = 0;
    }

    d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_DEFAULT;
    d3dpp.hDeviceWindow = hWnd;
    d3dpp.EnableAutoDepthStencil = TRUE;
    d3dpp.AutoDepthStencilFormat = D3DFMT_D16;
    d3dpp.Windowed = !fscreen;
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;

    SetWindowLong(hWnd, GWL_STYLE, WindowStyle);

    Cleanup();
    g_pd3dDevice->Reset(&d3dpp);
    ReInit();

    if (!fscreen) {
        SetWindowPos(hWnd, HWND_NOTOPMOST, 0, 0, WindowRect.right - WindowRect.left, WindowRect.bottom - WindowRect.top, SWP_SHOWWINDOW);
        CenterWindow(hWnd);
    }

    return true;
}

bool AllInit() {
    if (!InitD3D(Width, Height, fullscreen))
        return false;

    if (FAILED(g_pd3dDevice->CreateVertexBuffer(
        36 * sizeof(MyVertex),
        D3DUSAGE_WRITEONLY,
        MyVertex_FVF,
        D3DPOOL_MANAGED,
        &MyVertexBuffer,
        NULL)))
        return false;

    //   
    MyVertex v[] =
    {
        { -1.0f, -1.0f,  1.0f, D3DCOLOR_XRGB(255,   0,   0), 0.0f, 0.0f,},
        {  1.0f, -1.0f,  1.0f, D3DCOLOR_XRGB(255,   0,   0), 1.0f, 0.0f,},
        {  1.0f,  1.0f,  1.0f, D3DCOLOR_XRGB(255,   0,   0), 1.0f, 1.0f,},
        { -1.0f, -1.0f,  1.0f, D3DCOLOR_XRGB(255,   0,   0), 0.0f, 0.0f,},
        {  1.0f,  1.0f,  1.0f, D3DCOLOR_XRGB(255,   0,   0), 1.0f, 1.0f,},
        { -1.0f,  1.0f,  1.0f, D3DCOLOR_XRGB(255,   0,   0), 0.0f, 1.0f,},

        { -1.0f, -1.0f, -1.0f, D3DCOLOR_XRGB(0, 255,   0), 1.0f, 0.0f,},
        { -1.0f,  1.0f, -1.0f, D3DCOLOR_XRGB(0, 255,   0), 1.0f, 1.0f,},
        {  1.0f,  1.0f, -1.0f, D3DCOLOR_XRGB(0, 255,   0), 0.0f, 1.0f,},
        { -1.0f, -1.0f, -1.0f, D3DCOLOR_XRGB(0, 255,   0), 1.0f, 0.0f,},
        {  1.0f,  1.0f, -1.0f, D3DCOLOR_XRGB(0, 255,   0), 0.0f, 1.0f,},
        {  1.0f, -1.0f, -1.0f, D3DCOLOR_XRGB(0, 255,   0), 0.0f, 0.0f,},

        { -1.0f,  1.0f, -1.0f, D3DCOLOR_XRGB(0,   0, 255), 0.0f, 1.0f,},
        { -1.0f,  1.0f,  1.0f, D3DCOLOR_XRGB(0,   0, 255), 0.0f, 0.0f,},
        {  1.0f,  1.0f,  1.0f, D3DCOLOR_XRGB(0,   0, 255), 1.0f, 0.0f,},
        { -1.0f,  1.0f, -1.0f, D3DCOLOR_XRGB(0,   0, 255), 0.0f, 1.0f,},
        {  1.0f,  1.0f,  1.0f, D3DCOLOR_XRGB(0,   0, 255), 1.0f, 0.0f,},
        {  1.0f,  1.0f, -1.0f, D3DCOLOR_XRGB(0,   0, 255), 1.0f, 1.0f,},

        { -1.0f, -1.0f, -1.0f, D3DCOLOR_XRGB(255, 255,   0), 1.0f, 1.0f,},
        {  1.0f, -1.0f, -1.0f, D3DCOLOR_XRGB(255, 255,   0), 0.0f, 1.0f,},
        {  1.0f, -1.0f,  1.0f, D3DCOLOR_XRGB(255, 255,   0), 0.0f, 0.0f,},
        { -1.0f, -1.0f, -1.0f, D3DCOLOR_XRGB(255, 255,   0), 1.0f, 1.0f,},
        {  1.0f, -1.0f,  1.0f, D3DCOLOR_XRGB(255, 255,   0), 0.0f, 0.0f,},
        { -1.0f, -1.0f,  1.0f, D3DCOLOR_XRGB(255, 255,   0), 1.0f, 0.0f,},

        {  1.0f, -1.0f, -1.0f, D3DCOLOR_XRGB(255,   0, 255), 1.0f, 0.0f,},
        {  1.0f,  1.0f, -1.0f, D3DCOLOR_XRGB(255,   0, 255), 1.0f, 1.0f,},
        {  1.0f,  1.0f,  1.0f, D3DCOLOR_XRGB(255,   0, 255), 0.0f, 1.0f,},
        {  1.0f, -1.0f, -1.0f, D3DCOLOR_XRGB(255,   0, 255), 1.0f, 0.0f,},
        {  1.0f,  1.0f,  1.0f, D3DCOLOR_XRGB(255,   0, 255), 0.0f, 1.0f,},
        {  1.0f, -1.0f,  1.0f, D3DCOLOR_XRGB(255,   0, 255), 0.0f, 0.0f,},

        { -1.0f, -1.0f, -1.0f, D3DCOLOR_XRGB(255, 255, 255), 0.0f, 0.0f,},
        { -1.0f, -1.0f,  1.0f, D3DCOLOR_XRGB(255, 255, 255), 1.0f, 0.0f,},
        { -1.0f,  1.0f,  1.0f, D3DCOLOR_XRGB(255, 255, 255), 1.0f, 1.0f,},
        { -1.0f, -1.0f, -1.0f, D3DCOLOR_XRGB(255, 255, 255), 0.0f, 0.0f,},
        { -1.0f,  1.0f,  1.0f, D3DCOLOR_XRGB(255, 255, 255), 1.0f, 1.0f,},
        { -1.0f,  1.0f, -1.0f, D3DCOLOR_XRGB(255, 255, 255), 0.0f, 1.0f,}
    };


    void* pBuf;
    MyVertexBuffer->Lock(0, 36 * sizeof(MyVertex), &pBuf, 0);
    memcpy(pBuf, v, 36 * sizeof(MyVertex));
    MyVertexBuffer->Unlock();

    D3DXCreateTextureFromFile(g_pd3dDevice, L"Texture.bmp", &MyTexture);

    ReInit();
    return true;
}

void AllShutdown() {
    Cleanup();

    SAFE_RELEASE(MyTexture);
    SAFE_RELEASE(MyVertexBuffer);
    SAFE_RELEASE(g_pd3dDevice);
    SAFE_RELEASE(g_pD3D);
}

void Render() {
    HRESULT hr;

    hr = g_pd3dDevice->TestCooperativeLevel();

    if (hr == D3DERR_DEVICELOST)
        return;

    if (hr == D3DERR_DEVICENOTRESET) {
        Cleanup();
        g_pd3dDevice->Reset(&d3dpp);
        if (g_pd3dDevice)
            ReInit();
    }

    if (NULL == g_pd3dDevice)
        return;

    if (SUCCEEDED(g_pd3dDevice->BeginScene())) {
        float time = GetTickCount() / 2048.0f;

        g_pd3dDevice->Clear(0L, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(
            int((cos(time * 2.0f) + 1.0f) / 2.0f * 255),
            int((cos(time) + 1.0f) / 2.0f * 255),
            int((sin(time * 2.0f) + 1.0f) / 2.0f * 255)), 1.0f, 0L);

        D3DXMATRIX matWorld;
        D3DXMatrixRotationYawPitchRoll(&matWorld,
            GetTickCount() / 1024.0f, GetTickCount() / 2048.0f, GetTickCount() / 2048.0f);

        g_pd3dDevice->SetTransform(D3DTS_WORLD, &matWorld);

        D3DXMATRIX matView;
        D3DXVECTOR3 eye(4.0f, 4.0f, 4.0f);
        D3DXVECTOR3 at(0.0f, 0.0f, 0.0f);
        D3DXVECTOR3 up(0.0f, 1.0f, 0.0f);

        D3DXMatrixLookAtLH(&matView, &eye, &at, &up);
        g_pd3dDevice->SetTransform(D3DTS_VIEW, &matView);

        g_pd3dDevice->SetStreamSource(0, MyVertexBuffer, 0, sizeof(MyVertex));
        g_pd3dDevice->SetFVF(MyVertex_FVF);
        g_pd3dDevice->SetTexture(0, MyTexture);
        g_pd3dDevice->DrawPrimitive(D3DPT_TRIANGLELIST, 0, 12);

        g_pd3dDevice->EndScene();
    }

    g_pd3dDevice->Present(NULL, NULL, NULL, NULL);
}

LRESULT CALLBACK WinProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) {
    switch (message) {
    case WM_SYSCOMMAND:
        if (wParam == SC_SIZE && !fullscreen) {
            return 0;
        }
        break;

    case WM_CLOSE:
        isProcess = false;
        PostQuitMessage(0);
        break;

    case WM_KEYDOWN:
        keys[wParam] = true;
        break;

    case WM_KEYUP:
        keys[wParam] = false;
        break;

    default:
        return DefWindowProc(hWnd, message, wParam, lParam);
    }

    return 0;
}

bool RegisterWindowClass(HINSTANCE hInstance) {
    WNDCLASSEX wc;
    wc.cbSize = sizeof(WNDCLASSEX);
    wc.style = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc = (WNDPROC)WinProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = hInstance;
    wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
    wc.lpszMenuName = NULL;
    wc.lpszClassName = CLASSNAME;
    wc.hIconSm = LoadIcon(NULL, IDI_APPLICATION);

    if (!RegisterClassEx(&wc)) {
        MessageBox(NULL, L"    !", L"", MB_OK | MB_ICONEXCLAMATION);
        return false;
    }

    return true;
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow) {
    if (!RegisterWindowClass(hInstance))
        return 0;

    DWORD WindowStyle, WindowExStyle;

    if (fullscreen) {
        WindowExStyle = WS_EX_APPWINDOW;
        WindowStyle = WS_POPUP;
    }
    else {
        WindowExStyle = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
        WindowStyle = WS_OVERLAPPEDWINDOW;
    }

    WindowRect.left = (long)0;
    WindowRect.right = (long)Width;
    WindowRect.top = (long)0;
    WindowRect.bottom = (long)Height;

    AdjustWindowRectEx(&WindowRect, WindowStyle, FALSE, WindowExStyle);

    hWnd = CreateWindowEx(
        NULL,
        CLASSNAME,
        APPNAME,
        WindowStyle | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
        0, 0,
        WindowRect.right - WindowRect.left,
        WindowRect.bottom - WindowRect.top,
        NULL, NULL,
        hInstance, NULL);

    if (!hWnd) {
        MessageBox(NULL, L"   !", L"", MB_OK | MB_ICONEXCLAMATION);
        return 0;
    }

    //  
    CenterWindow(hWnd);

    ShowWindow(hWnd, nCmdShow);
    UpdateWindow(hWnd);
    SetFocus(hWnd);


    if (!AllInit())
        return 0;

    MSG msg;
    ZeroMemory(&msg, sizeof(msg));

    while (isProcess) {
        if (PeekMessage(&msg, NULL, 0U, 0U, PM_REMOVE)) {
            if (msg.message == WM_QUIT) {
                isProcess = false;
                break;
            }

            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
        else {
            if (keys[VK_ESCAPE]) {
                isProcess = false;
                break;
            }

            if (keys[(int)'F']) {
                keys[(int)'F'] = false;
                fullscreen = !fullscreen;
                ResetWindow(Width, Height, fullscreen);
            }

            Render();
        }
    }

    AllShutdown();

    return (int)msg.wParam;
}
